#include <math.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

#define frag_size 2048 /* bytes   */
#define samp_size 1024 /* samples */

volatile int adevice,diskwriter;
int arg,status,caps;
pthread_t reader_id, writer_id, processor_id;
pthread_attr_t attr;
struct timeval tv;
fd_set rfds;
unsigned char ichar;
pthread_mutex_t imutex,omutex;

struct sbuffer  {       volatile struct sbuffer *next;
                        volatile signed short int data[samp_size];
                };

volatile struct sbuffer *outroot=NULL, *inroot=NULL;

volatile int SLOOP_SHUTDOWN=0, SLOOP_EXIT=0, SLOOP_DISK=0, SLOOP_IMEM=0, SLOOP_OMEM=0, SLOOP_SLEEP=0;





struct sbuffer *switch_frag ( struct sbuffer *ptr )
{
        struct sbuffer *help;
        int x;
        long int sum;

        if(ptr)
        {       ptr->next=NULL;
                pthread_mutex_lock(&omutex);
                        if(outroot)
                        {
                                help=(struct sbuffer *)outroot; SLOOP_OMEM=1;
                                while(help->next) { help=(struct sbuffer *)help->next; SLOOP_OMEM++; }
                                help->next=ptr;
                        }
                        else { outroot=ptr; SLOOP_OMEM=1; }
                pthread_mutex_unlock(&omutex);

                ptr=(struct sbuffer * ) malloc(sizeof(struct sbuffer));
                if(ptr==NULL) { SLOOP_SHUTDOWN=2; printf("OUT OF SLOOP_PROCESSOR MEMORY\n"); pthread_exit(NULL); }
        }
        else
        {
                pthread_mutex_lock(&omutex);
                        if(outroot)
                        {
                                help=(struct sbuffer *)outroot; SLOOP_OMEM=1;
                                while(help->next) { ptr=help; help=(struct sbuffer *)help->next; SLOOP_OMEM++; }

                                if (SLOOP_OMEM>1)
                                {       for (sum=x=0;x<samp_size;x++) sum+=abs(help->data[x]); sum/=samp_size;
                                        if (sum<16) { ptr->next=NULL; free((void *)help); }
                                }
                        }
                        else { SLOOP_OMEM=0; }
                pthread_mutex_unlock(&omutex);
        }

        return(ptr);
}





void * reader (void * dummy)
{
        struct sbuffer *ptr,*help;
        int st;

        while (SLOOP_SHUTDOWN==0)
        {
                if(!SLOOP_SLEEP)
                {
                        ptr=(struct sbuffer * ) malloc(sizeof(struct sbuffer));
                        if(ptr==NULL) { SLOOP_SHUTDOWN=2; printf("OUT OF SLOOP_READER MEMORY\n"); pthread_exit(NULL); }
                        st=read(adevice, (void *)&(ptr->data[0]), frag_size);
                        if(st!=frag_size) { SLOOP_SHUTDOWN=2; printf("READER ERROR\n"); pthread_exit(NULL); }

                        ptr->next=NULL;

                        pthread_mutex_lock(&imutex);
                                if (inroot==NULL) { inroot=ptr; SLOOP_IMEM=1; }
                                else
                                {       help=(struct sbuffer *)inroot; SLOOP_IMEM=1;
                                        while(help->next) { help=(struct sbuffer *)help->next; SLOOP_IMEM++; }
                                        help->next=ptr;
                                }
                        pthread_mutex_unlock(&imutex);
                }
                else
                {
                        if(SLOOP_IMEM)
                        {       pthread_mutex_lock(&imutex);
                                help=(struct sbuffer *)inroot;
                                while (help) { ptr=help; help=(struct sbuffer *)help->next; free((void *)ptr); }
                                inroot=NULL;
                                SLOOP_IMEM=0;
                                pthread_mutex_unlock(&imutex);
                        }
                }
        }

        if (!(--SLOOP_SHUTDOWN)) SLOOP_EXIT=1;
        pthread_exit(NULL);
}





void * processor (void * dummy)
{
        struct sbuffer *ptr;
        long int a,c,x,y,z,n,temp,temp2;
        long int period,lperiod,multi;

        int found[samp_size*2][3];
        int workptr=0,fragptr=0,walker=0,imem,omem;
        signed short int worker[samp_size*3],wrapper[samp_size];

        lperiod=period=0; temp=0;

        while (SLOOP_SHUTDOWN==0)
        {
                pthread_mutex_lock(&imutex); imem=SLOOP_IMEM; pthread_mutex_unlock(&imutex);
                pthread_mutex_lock(&omutex); omem=SLOOP_OMEM; pthread_mutex_unlock(&omutex);

                if( ((SLOOP_SLEEP)||(period<lperiod)) && ((imem+omem)>88) )
                {
                        switch_frag(NULL);
                        SLOOP_SLEEP=1; continue;
                }
                else if (SLOOP_SLEEP)
                {
                        SLOOP_SLEEP=0; workptr=0;
                        if(fragptr) { for(x=0;x<fragptr;x++) wrapper[x]=(wrapper[x]*(fragptr-x))/fragptr; }
                        do {pthread_mutex_lock(&imutex); ptr=(struct sbuffer *)inroot; pthread_mutex_unlock(&imutex);} while(!ptr);
                        for(x=0;x<32;x++) inroot->data[x]=(inroot->data[x]*x)/32;
                }

                lperiod=period;
                        pthread_mutex_lock(&imutex);

                if(inroot)
                {
                        ptr=(struct sbuffer *)inroot;
                        if (inroot->next) inroot=inroot->next;
                        else inroot=NULL;

                        pthread_mutex_unlock(&imutex);

                        memcpy((void *)&(worker[workptr]), (void *)&(ptr->data[0]), frag_size);
                        workptr+=samp_size; found[0][0]=0; n=0; z=0;
                        if (fragptr)
                                memcpy( (void *)&(ptr->data[0]) , (void *)&wrapper[0] , fragptr*2);

                        for(x=1;x<(workptr-1);x++)
                        {
                                if (    ( (worker[x-1]>0) && (worker[x-1]>worker[x]) && (worker[x]>worker[x+1]) && (worker[x+1]<0) )||
                                        ( (worker[x-1]<0) && (worker[x-1]<worker[x]) && (worker[x]<worker[x+1]) && (worker[x+1]>0) )
                                   )
                                {
                                        switch (n)
                                        {
                                                case 2:
                                                {       if ( ( ((x-found[z][1])<=441) && ((x-found[z][1])>=6) ) && ( (x-found[z][1])>=((found[z][1]-found[z][0])/8) ) )
                                                        {       temp2=0; for (y=found[z][1]; y<x; y++) temp2=temp2+worker[y];
                                                                if ( ( abs(temp2/(x-found[z][1])) > 512 ) && ( (temp2/abs(temp2))!=(temp/abs(temp)) )  )
                                                                { found[z][n]=x; n++; break; }
                                                                else
                                                                { n=1; }
                                                        }
                                                        else    { n=1; }
                                                }

                                                case 1:
                                                {       if ( ((x-found[z][0])<=441) && ((x-found[z][0])>=6) )
                                                        {       temp=0; for (y=found[z][0]; y<x; y++) temp=temp+worker[y];
                                                                if ( abs(temp/(x-found[z][0])) > 512 )
                                                                { found[z][n]=x; n++; break; }
                                                                else
                                                                { n=0; }
                                                        }
                                                        else    { n=0; }
                                                }

                                                case 0:
                                                {       found[z][n]=x;
                                                        n++;
                                                        break;
                                                }

                                                default:
                                                        break;
                                        }

                                        if (n==3)
                                        {
                                                found[z+1][0]=found[z][2]; z++; n=1;
                                        }
                                }
                        }

                        walker=0;

                        if (z==0)
                        {       memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[0] , (samp_size-fragptr)*2 );
                                ptr=switch_frag(ptr);
                                walker=samp_size-fragptr;
                                fragptr=0;
                        }
                        else
                        {
                                for(x=0;x<z;x++)
                                {
                                        c=(found[x][0]-walker);
                                        if (c>0)
                                        {
                                                while (c>(y=samp_size-fragptr))
                                                {
                                                        memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[walker]  , y*2 );
                                                        ptr=switch_frag(ptr);
                                                        fragptr=0; c-=y; walker+=y;
                                                }

                                                memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[walker]  , c*2 );
                                                fragptr+=c; walker+=c;
                                                if (fragptr==samp_size) { ptr=switch_frag(ptr); fragptr=0; }
                                        }

                                        period++; period=period%(3000); multi=labs(1500-period);

                                        multi=(long int) pow((((double)multi)/750.0),(10.0*((double)multi))/1000.0);
                                        multi+=(labs(1500-period)/1000)+1;

                                        c=(found[x][2]-found[x][0]);
                                        for(a=0;a<=multi/c;a++)
                                        {
                                                y=samp_size-fragptr;
                                                if(c>y)
                                                {
                                                        memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[found[x][0]]   , (y)*2 );
                                                        ptr=switch_frag(ptr);
                                                        memcpy( (void *)&(ptr->data[0]) ,       (void *)&worker[found[x][0]+y] , (c-y)*2  );
                                                        fragptr=(c-y);
                                                }
                                                else
                                                {
                                                        memcpy( (void *)&(ptr->data[fragptr]) , (void *)&worker[found[x][0]]   , c*2 );
                                                        fragptr+=c;
                                                        if (fragptr==samp_size) { ptr=switch_frag(ptr); fragptr=0; }
                                                }
                                        }
                                        walker=found[x][2];
                                }
                        }

                        x=workptr-walker;
                        if (x>0)
                        {
                                memmove( (void *)&worker[0] , (void *)&worker[walker], x*2 );
                                workptr=x;
                        }
                        else    workptr=0;

                        if (fragptr) memcpy( (void *)&wrapper[0] , (void *)&(ptr->data[0]) , fragptr*2); free((void *)ptr);

                }
                else
                {       pthread_mutex_unlock(&imutex);  }

        }

        if (!(--SLOOP_SHUTDOWN)) SLOOP_EXIT=1;
        pthread_exit(NULL);
}





void * writer (void * dummy)
{
        struct sbuffer *ptr;
        int st,disk_counter=0;

        while (SLOOP_SHUTDOWN==0)
        {
                        pthread_mutex_lock(&omutex);

                if (outroot)
                {
                        ptr=(struct sbuffer *)outroot;
                        if (outroot->next) outroot=outroot->next;
                        else outroot=NULL;

                        pthread_mutex_unlock(&omutex);

                        st=write(adevice, (void *)&(ptr->data[0]), frag_size);
                        if(st!=frag_size) { SLOOP_SHUTDOWN=2; printf("WRITER ERROR\n"); pthread_exit(NULL); }
                        if (SLOOP_DISK)
                                {       st=write(diskwriter, (void *)&(ptr->data[0]), frag_size);
                                        if(st!=frag_size)
                                        {       printf("ERROR while writing output file. Closing.\n");
                                                if (close(diskwriter)) printf("Closing output file failed.\n");
                                                SLOOP_DISK=0;
                                        }
                                        else
                                        {       disk_counter++;
                                                if (disk_counter>=1048576)
                                                {       printf("Maximum file size reached.\n");
                                                        if (close(diskwriter))  printf("Closing output file failed.\n");
                                                        else                    printf("Closing output file.\n");
                                                        SLOOP_DISK=0;
                                                }
                                        }
                                }

                        free((void *)ptr);
                }
                else
                {       pthread_mutex_unlock(&omutex);  }
        }

        if (!(--SLOOP_SHUTDOWN)) SLOOP_EXIT=1;
        pthread_exit(NULL);
}





int main (int argc, char** argv)
{
        struct sbuffer *ptr, *help;
        int x,y,z;

        char roto[] = "\\|/-";
        char *ptxt[]  = {       "      [recording]    \n",
                                "  [recording  paused]\n"
                        };


        printf("\nsloop.splitter\n\n");

        adevice = open ("/dev/dsp", O_RDWR|O_SYNC);
	if (adevice < 0)
        {       printf("Open of /dev/dsp failed.\n");
                exit(1);
        }
        printf("Audio device opened.\n");

        ioctl(adevice, SNDCTL_DSP_SETDUPLEX, 0);
        ioctl(adevice, SNDCTL_DSP_GETCAPS  , &caps);
        if (caps&DSP_CAP_DUPLEX)
        {       printf("Fullduplex: Yes.\n");
        }
        else
        {       printf("Fullduplex: No.\n");
                close(adevice);
                exit(1);
        }

        arg = AFMT_S16_LE; /* sample size */
	status = ioctl(adevice, SOUND_PCM_WRITE_BITS, &arg);
	if (status == -1)
        {       printf("SOUND_PCM_WRITE_BITS ioctl failed.\n");
                close(adevice);
                exit(1);
        }
	if (arg != AFMT_S16_LE)
        {       printf("Unable to set sample size.\n");
                close(adevice);
                exit(1);
        }

	arg = 1;  /* mono or stereo */
	status = ioctl(adevice, SOUND_PCM_WRITE_CHANNELS, &arg);
	if (status == -1)
        {       printf("SOUND_PCM_WRITE_CHANNELS ioctl failed.\n");
                close(adevice);
                exit(1);
        }
	if (arg != 1)
        {       printf("Unable to set number of channels.\n");
                close(adevice);
                exit(1);
        }

        arg = 44100;      /* sampling rate */
	status = ioctl(adevice, SOUND_PCM_WRITE_RATE, &arg);
	if (status == -1)
        {       printf("SOUND_PCM_WRITE_WRITE ioctl failed.\n");
                close(adevice);
                exit(1);
        }

        printf("Device parameters set.\n");

        if (pthread_mutex_init(&imutex, NULL) < 0)
        {
                printf("Input mutex failed.\n");
                close(adevice);
                exit(1);
        }

        if (pthread_mutex_init(&omutex, NULL) < 0)
        {
                printf("Output mutex failed.\n");
                close(adevice);
                exit(1);
        }

        if (argc==3)
        {       if (!strcmp(argv[1],"-o"))
                {       diskwriter=open(argv[2],O_WRONLY|O_CREAT|O_TRUNC );
                        if (diskwriter==-1)     printf("Creating ouput file %s failed.\n",argv[2]);
                        else
                        {       SLOOP_DISK=1;
                                printf("Creating ouput file %s\n",argv[2]);
                                if (fchmod(diskwriter,S_IWUSR|S_IRUSR)) printf("Cannot set file permissions.\n");
                        }

                }
        }

        pthread_attr_init(&attr);
        pthread_create(&reader_id,   &attr, reader,   NULL);
        pthread_create(&writer_id,   &attr, writer,   NULL);
        pthread_create(&processor_id,&attr, processor,NULL);

        printf("Threads started.\nEntering main loop.\n\nHit 'q' <enter> to quit.\n\n\n\n\n\n\n"); z=0;

        do {
                FD_ZERO(&rfds);
                FD_SET(STDIN_FILENO, &rfds);
                tv.tv_sec=1; tv.tv_usec=0;
                ichar=0;
                if ( select(1, &rfds,NULL,NULL,&tv) ) { ichar=getchar(); }
                pthread_mutex_lock(&imutex); x=SLOOP_IMEM; pthread_mutex_unlock(&imutex);
                pthread_mutex_lock(&omutex); y=SLOOP_OMEM; pthread_mutex_unlock(&omutex);
                if ( ichar!='q' && !SLOOP_EXIT )
                {       printf("\033[5A\rInputQueue :%5d %cByte\nOutputQueue:%5d %cByte\nTotal      :%5d %cByte\n%s\n    \r(%c) [", x>512?x/512:x*2, x>512?'M':'k', y>512?y/512:y*2, y>512?'M':'k', (y+x)>512?(y+x)/512:(y+x)*2, (y+x)>512?'M':'k',ptxt[SLOOP_SLEEP], roto[(z++)%4]);
                        fflush(stdout);
                }
                else break;
        } while (1);

        printf("\nExiting ... please wait     "); fflush(stdout);

        if(!SLOOP_EXIT)
        {
                SLOOP_SHUTDOWN=3;x=0;
                while(SLOOP_SHUTDOWN)   {       printf("\b\b\b\b(%c) ",roto[(++x)%4]); fflush(stdout);
                                                tv.tv_sec=0; tv.tv_usec=100000;
                                                select(0,NULL,NULL,NULL,&tv);
                                        }
        }

        printf("\n"); fflush(stdout);

        ioctl(adevice, SNDCTL_DSP_RESET, 0);
        if (close(adevice))     printf("Closing audio device failed.\n");
        else                    printf("Closing audio device.\n");

        if (SLOOP_DISK)
        {       if (close(diskwriter))  printf("Closing output file failed.\n");
                else                    printf("Closing output file.\n");
        }

        printf("Cleaning up memory...\n"); fflush(stdout);

        help=(struct sbuffer *)inroot;
        while (help) { ptr=help; help=(struct sbuffer *)help->next; free((void *)ptr); }
        help=(struct sbuffer *)outroot;
        while (help) { ptr=help; help=(struct sbuffer *)help->next; free((void *)ptr); }

        printf("\n");
        return 0;

}
